;;; Directory Local Variables         -*- no-byte-compile: t; -*-
;;; For more information see (info "(emacs) Directory Variables")

((nil . ((tab-width . 8)
         (fill-column . 76)
         (emacs-lisp-docstring-fill-column . 76)))
 (emacs-lisp-mode . ((indent-tabs-mode . nil)))
 (lisp-data-mode . ((indent-tabs-mode . nil))))

